#define STRICT
#define VC_EXTRALEAN
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stddef.h>
#include <strsafe.h>
#include "resource.h"
#pragma check_stack(off)
#pragma runtime_checks( "", off )
#pragma intrinsic(memset)
#pragma intrinsic(memcmp)
#pragma intrinsic(memcpy)

#pragma comment(lib,"version.lib")

static HINSTANCE hInst;

static BOOL (WINAPI *_SfcIsFileProtected)(IN HANDLE RpcHandle,IN LPCWSTR ProtFileName) = NULL;

#define STRING_LENGTH(s) (sizeof(s)/sizeof(s[0]))

static LCID Uni_our_locale = 0;
static UINT Uni_our_codepage = 0;
static bool Uni_is_NT = false;

static HANDLE (WINAPI *_FindFirstFileW)(IN LPCWSTR lpFileName,OUT LPWIN32_FIND_DATAW lpFindFileData) = NULL;
static BOOL (WINAPI *_FindNextFileW)(IN HANDLE hFindFile,OUT LPWIN32_FIND_DATAW lpFindFileData) = NULL;
static HANDLE (WINAPI *_CreateFileW)(IN LPCWSTR lpFileName,IN DWORD dwDesiredAccess,IN DWORD dwShareMode,IN LPSECURITY_ATTRIBUTES lpSecurityAttributes,IN DWORD dwCreationDisposition,IN DWORD dwFlagsAndAttributes,IN HANDLE hTemplateFile) = NULL;
static BOOL (WINAPI *_SetFileAttributesW)(IN LPCWSTR lpFileName,IN DWORD dwFileAttributes) = NULL;

static DWORD (WINAPI *_GetFileVersionInfoSizeW)(LPWSTR lptstrFilename,LPDWORD lpdwHandle) = NULL;
static BOOL (WINAPI *_GetFileVersionInfoW)(LPWSTR lptstrFilename,DWORD dwHandle,DWORD dwLen,LPVOID lpData) = NULL;

static HANDLE Uni_FindFirstFile(IN LPCWSTR lpFileName,OUT LPWIN32_FIND_DATAW lpFindFileData)
{
	if (_FindFirstFileW)
	{
		return _FindFirstFileW(lpFileName,lpFindFileData);
	}
	else
	{
		char temp[MAX_PATH];
		WideCharToMultiByte(Uni_our_codepage,0,lpFileName,-1,temp,STRING_LENGTH(temp),NULL,NULL);
		WIN32_FIND_DATAA fd;
		HANDLE res = FindFirstFileA(temp,&fd);
		if (res)
		{
			memcpy(lpFindFileData,&fd,offsetof(WIN32_FIND_DATAA,cFileName));
			MultiByteToWideChar(Uni_our_codepage,0,fd.cFileName,-1,lpFindFileData->cFileName,STRING_LENGTH(lpFindFileData->cFileName));
			MultiByteToWideChar(Uni_our_codepage,0,fd.cAlternateFileName,-1,lpFindFileData->cAlternateFileName,STRING_LENGTH(lpFindFileData->cAlternateFileName));
		}
		return res;
	}
}

static BOOL Uni_FindNextFile(IN HANDLE hFindFile,OUT LPWIN32_FIND_DATAW lpFindFileData)
{
	if (_FindFirstFileW)
	{
		return _FindNextFileW(hFindFile,lpFindFileData);
	}
	else
	{
		WIN32_FIND_DATAA fd;
		BOOL res = FindNextFileA(hFindFile,&fd);
		if (res)
		{
			memcpy(lpFindFileData,&fd,sizeof(DWORD)*5 + sizeof(FILETIME)*3);
			MultiByteToWideChar(Uni_our_codepage,0,fd.cFileName,-1,lpFindFileData->cFileName,STRING_LENGTH(lpFindFileData->cFileName));
			MultiByteToWideChar(Uni_our_codepage,0,fd.cAlternateFileName,-1,lpFindFileData->cAlternateFileName,STRING_LENGTH(lpFindFileData->cAlternateFileName));
		}
		return res;
	}
}

static HANDLE Uni_CreateFile(IN LPCWSTR lpFileName,IN DWORD dwDesiredAccess,IN DWORD dwShareMode,IN LPSECURITY_ATTRIBUTES lpSecurityAttributes,IN DWORD dwCreationDisposition,IN DWORD dwFlagsAndAttributes,IN HANDLE hTemplateFile)
{
	if (_CreateFileW)
	{
		return _CreateFileW(lpFileName,dwDesiredAccess,dwShareMode,lpSecurityAttributes,dwCreationDisposition,dwFlagsAndAttributes,hTemplateFile);
	}
	else
	{
		char temp[MAX_PATH];
		WideCharToMultiByte(Uni_our_codepage,0,lpFileName,-1,temp,STRING_LENGTH(temp),NULL,NULL);
		return CreateFileA(temp,dwDesiredAccess,dwShareMode,lpSecurityAttributes,dwCreationDisposition,dwFlagsAndAttributes,hTemplateFile);
	}
}

static BOOL Uni_SetFileAttributes(IN LPCWSTR lpFileName,IN DWORD dwFileAttributes)
{
	if (_SetFileAttributesW)
	{
		return _SetFileAttributesW(lpFileName,dwFileAttributes);
	}
	else
	{
		char temp[MAX_PATH];
		WideCharToMultiByte(Uni_our_codepage,0,lpFileName,-1,temp,STRING_LENGTH(temp),NULL,NULL);
		return SetFileAttributesA(temp,dwFileAttributes);
	}
}

static DWORD Uni_GetFileVersionInfoSize(LPWSTR lptstrFilename,LPDWORD lpdwHandle)
{
	if (_GetFileVersionInfoSizeW)
	{
		return _GetFileVersionInfoSizeW(lptstrFilename,lpdwHandle);
	}
	else
	{
		char temp[MAX_PATH];
		WideCharToMultiByte(Uni_our_codepage,0,lptstrFilename,-1,temp,STRING_LENGTH(temp),NULL,NULL);
		return GetFileVersionInfoSizeA(temp,lpdwHandle);
	}
}

static BOOL Uni_GetFileVersionInfo(LPWSTR lptstrFilename,DWORD dwHandle,DWORD dwLen,LPVOID lpData)
{
	if (_GetFileVersionInfoW)
	{
		return _GetFileVersionInfoW(lptstrFilename,dwHandle,dwLen,lpData);
	}
	else
	{
		char temp[MAX_PATH];
		WideCharToMultiByte(Uni_our_codepage,0,lptstrFilename,-1,temp,STRING_LENGTH(temp),NULL,NULL);
		return GetFileVersionInfoA(temp,dwHandle,dwLen,lpData);
	}
}

static int Uni_strcmp(LPCWSTR lpString1,LPCWSTR lpString2)
{
	while (1)
	{
		WCHAR c1 = *lpString1++;
		WCHAR c2 = *lpString2++;
		if (c1<c2)
			return -1;
		if (c1>c2)
			return 1;
		if (c1==0)
			return 0;
	}
}

static int Uni_strcmpi(LPCWSTR lpString1,LPCWSTR lpString2)
{
	// not international safe, but good enough for this use :-)
	while (1)
	{
		WCHAR c1 = *lpString1++;
		WCHAR c2 = *lpString2++;
		if (c1>='A' && c1<='Z')
			c1 = (WCHAR)(c1-'A'+'a');
		if (c2>='A' && c2<='Z')
			c2 = (WCHAR)(c2-'A'+'a');
		if (c1<c2)
			return -1;
		if (c1>c2)
			return 1;
		if (c1==0)
			return 0;
	}
}

static void Uni_strcpyn(LPWSTR dest,LPCWSTR source,int count)
{
	WCHAR *dest_end = dest+count-1;
	while (dest<dest_end)
	{
		WCHAR c = *source++;
		if (!c)
			break;
		*dest++ = c;
	}
	*dest = 0;
}

static void Uni_strcatn(LPWSTR dest,LPCWSTR source,int count)
{
	WCHAR *dest_end = dest+count-1;
	while (*dest && dest<dest_end)
		dest++;
	while (dest<dest_end)
	{
		WCHAR c = *source++;
		if (!c)
			break;
		*dest++ = c;
	}
	*dest = 0;
}

static int Uni_strlen(LPCWSTR lpString1)
{
	LPCWSTR p = lpString1;
	while (*p)
		p++;
	return (int)(p-lpString1);
}

static void Uni_Init(void)
{
	Uni_is_NT = !(GetVersion() & 0x80000000);
	if (Uni_is_NT)
	{
		HINSTANCE kernel32 = GetModuleHandle("kernel32");
		*((FARPROC*)(&_FindFirstFileW)) = GetProcAddress(kernel32,"FindFirstFileW");
		*((FARPROC*)(&_FindNextFileW)) = GetProcAddress(kernel32,"FindNextFileW");
		*((FARPROC*)(&_CreateFileW)) = GetProcAddress(kernel32,"CreateFileW");
		*((FARPROC*)(&_SetFileAttributesW)) = GetProcAddress(kernel32,"SetFileAttributesW");
		HINSTANCE version = GetModuleHandle("version");
		*((FARPROC*)(&_GetFileVersionInfoSizeW)) = GetProcAddress(version,"GetFileVersionInfoSizeW");
		*((FARPROC*)(&_GetFileVersionInfoW)) = GetProcAddress(version,"GetFileVersionInfoW");
	}
	Uni_our_locale = GetUserDefaultLCID();
	char code_page_name[6];
	GetLocaleInfo(Uni_our_locale,LOCALE_IDEFAULTANSICODEPAGE,code_page_name,sizeof(code_page_name));
	Uni_our_codepage = 0;
	char *p = code_page_name;
	while (*p)
		Uni_our_codepage = Uni_our_codepage*10 + ((*p++)-'0');
}

static __int64 GetFileVersion(const WCHAR *filename)
{
	__int64 ver = 0;
	DWORD temp;
	DWORD size = Uni_GetFileVersionInfoSize((WCHAR*)filename,&temp);
	if (size)
	{
		BYTE *buf = (BYTE *)LocalAlloc(LMEM_FIXED,size);
		if (buf)
		{
			Uni_GetFileVersionInfo((WCHAR*)filename,0,size,buf);
			VS_FIXEDFILEINFO *_ver = NULL;
			UINT foundsize;
			VerQueryValue(buf,"\\",(void**)&_ver,&foundsize);
			if (_ver && (foundsize>=16))
			{
				((DWORD*)&ver)[0] = _ver->dwFileVersionLS;
				((DWORD*)&ver)[1] = _ver->dwFileVersionMS;
			}
			LocalFree(buf);
		}
	}
	return ver;
}

static void __cdecl output(const WCHAR *lpszFormat, ...)
{
    WCHAR MsgBuff[4096];
    va_list argList;
    int nBuf;
    va_start(argList,lpszFormat);
	StringCchVPrintfW(MsgBuff,STRING_LENGTH(MsgBuff),lpszFormat,argList);
	nBuf = Uni_strlen(MsgBuff);
    va_end(argList);
    if (nBuf >= 0)
	{
		char temp[4096];
		WideCharToMultiByte(Uni_our_codepage,0,MsgBuff,-1,temp,STRING_LENGTH(temp),NULL,NULL);
		printf(temp);
	}
}

static void ProcessFolder(const WCHAR *foldername)
{
	WIN32_FIND_DATAW fdd;
	WCHAR s[MAX_PATH];
	Uni_strcpyn(s,foldername,STRING_LENGTH(s));
	Uni_strcatn(s,L"*.*",STRING_LENGTH(s));
	HANDLE hfind = Uni_FindFirstFile(s,&fdd);
	if (hfind!=INVALID_HANDLE_VALUE)
	{
		do {
			if (fdd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)
			{
				if (Uni_strcmp(fdd.cFileName,L".")!=0 && Uni_strcmp(fdd.cFileName,L"..")!=0)
				{
					Uni_strcpyn(s,foldername,STRING_LENGTH(s));
					Uni_strcatn(s,fdd.cFileName,STRING_LENGTH(s));
					Uni_strcatn(s,L"\\",STRING_LENGTH(s));
					ProcessFolder(s);
				}
			}
			else
			{
				if (Uni_strcmpi(fdd.cFileName,L"gdiplus.dll")==0)
				{
					Uni_strcpyn(s,foldername,STRING_LENGTH(s));
					Uni_strcatn(s,fdd.cFileName,STRING_LENGTH(s));
					__int64 filever = GetFileVersion(s);
#define MAKEVER(v1,v2,v3,v4) ( ((__int64)(v1)<<48) + ((__int64)(v2)<<32) + ((__int64)(v3)<<16) + ((__int64)(v4)<<0) )
					if (filever<MAKEVER(5,1,3102,1360))
#undef MAKEVER
					{
						output(L"\"%s\" needs replacing (current ver %i.%i.%i.%i)",s,(int)((filever>>48)&0xffff),(int)((filever>>32)&0xffff),(int)((filever>>16)&0xffff),(int)((filever>>0)&0xffff));
						if (_SfcIsFileProtected==NULL || _SfcIsFileProtected(NULL,s)==FALSE)
						{
							Uni_SetFileAttributes(s,FILE_ATTRIBUTE_NORMAL); // just incase there is a RO flag set
							HANDLE file = Uni_CreateFile(s,GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
							if (file!=INVALID_HANDLE_VALUE)
							{
								HRSRC res = FindResource(hInst,MAKEINTRESOURCE(IDR_GDIPLUS),"DATA");
								if (res)
								{
									DWORD temp;
									WriteFile(file,LockResource(LoadResource(hInst,res)),SizeofResource(hInst,res),&temp,NULL);
									SetEndOfFile(file);
									output(L" Replaced");
								}
								CloseHandle(file);
							}
							else
							{
								output(L" Failed to open (error:%i)",GetLastError());
							}
						}
						else
						{
							output(L" Skipping due to SystemFileProtection");
						}
						output(L"\r\n");
					}
				}
			}
		} while (Uni_FindNextFile(hfind,&fdd));
		FindClose(hfind);
	}
}

static void Run(void)
{
	DWORD drives = GetLogicalDrives();
	char driveA[4] = "A:\\";
	for (int i=0;i<26;++i)
	{
		if (drives&1)
		{
			switch (GetDriveType(driveA))
			{
			case DRIVE_FIXED:
				{
					WCHAR driveW[4];
					MultiByteToWideChar(Uni_our_codepage,0,driveA,-1,driveW,STRING_LENGTH(driveW));
					ProcessFolder(driveW);
				}
				break;
			}
		}
		drives >>= 1;
		driveA[0]++;
	}
}

int __cdecl main()
{
	hInst = GetModuleHandle(NULL);
	*((FARPROC*)(&_SfcIsFileProtected)) = GetProcAddress(LoadLibrary("sfc"),"SfcIsFileProtected");
	Uni_Init();
	Run();
}
